@@ -37,16 +37,14 @@ obj_print_data.vctrs_partial <- function(x, ...) {
 #' @rdname new_partial
 #' @export
 is_partial <- function(x) {
-  .Call(vctrs_is_partial, x)
+  .Call(ffi_is_partial, x)
 }
 
 #' @rdname new_partial
-#' @inheritParams ellipsis::dots_empty
+#' @inheritParams rlang::args_dots_empty
 #' @export
 vec_ptype_finalise <- function(x, ...) {
-  if (!missing(...)) {
-    ellipsis::check_dots_empty()
-  }
+  check_dots_empty0(...)
   return(.Call(vctrs_ptype_finalise, x))
   UseMethod("vec_ptype_finalise")
 }
