@@ -3,11 +3,8 @@
 
 test_that("data frames print nicely", {
   expect_equal(vec_ptype_abbr(mtcars), "df[,11]")
-
-  verify_output(test_path("test-type-data-frame.txt"), {
-    vec_ptype_show(mtcars)
-    vec_ptype_show(iris)
-  })
+  expect_snapshot(vec_ptype_show(mtcars))
+  expect_snapshot(vec_ptype_show(iris))
 })
 
 test_that("embedded data frames print nicely", {
@@ -16,9 +13,7 @@ test_that("embedded data frames print nicely", {
   df$b <- list_of(1, 2, 3)
   df$c <- as_list_of(split(data.frame(x = 1:3, y = letters[1:3]), 1:3))
 
-  verify_output(test_path("test-type-data-frame-embedded.txt"), {
-    vec_ptype_show(df)
-  })
+  expect_snapshot(vec_ptype_show(df))
 })
 
 # coercing ----------------------------------------------------------------
@@ -58,26 +53,13 @@ test_that("combining data frames with foreign classes uses fallback", {
   foo <- foobar(data.frame())
   df <- data.frame()
 
-  # Same type fallback
   expect_identical(vec_ptype_common(foo, foo, foo), foo)
-  expect_incompatible_df(vec_ptype_common(foo, foo, df, foo), df)
+  expect_identical(vec_ptype_common(foo, foo, df, foo), df)
 
-  expect_identical(
-    expect_df_fallback_warning(vec_ptype2_fallback(foo, df)),
-    new_fallback_df(df, c("vctrs_foobar", "data.frame"))
-  )
-  expect_identical(
-    expect_df_fallback_warning(vec_ptype2_fallback(df, foo)),
-    new_fallback_df(df, c("data.frame", "vctrs_foobar"))
-  )
-  expect_identical(
-    expect_df_fallback_warning(vec_ptype_common_df_fallback(foo, df)),
-    df
-  )
-  expect_identical(
-    expect_df_fallback_warning(vec_ptype_common_df_fallback(df, foo)),
-    df
-  )
+  expect_identical(vec_ptype2(foo, df), data.frame())
+  expect_identical(vec_ptype2(df, foo), data.frame())
+  expect_identical(vec_ptype_common(foo, df), df)
+  expect_identical(vec_ptype_common(df, foo), df)
 
   cnds <- list()
   withCallingHandlers(
@@ -86,41 +68,22 @@ test_that("combining data frames with foreign classes uses fallback", {
       invokeRestart("muffleWarning")
     },
     expect_identical(
-      vec_ptype_common_df_fallback(foo, df, foo, foo),
+      vec_ptype_common(foo, df, foo, foo),
       df
     )
   )
 
-  # There should be only one warning even if many fallbacks
-  expect_length(cnds, 1)
-  expect_s3_class(cnds[[1]], "warning")
-  expect_match(cnds[[1]]$message, "falling back to <data.frame>")
+  # There are no longer any warnings
+  expect_length(cnds, 0)
 
-  expect_incompatible_df(
+  expect_equal(
     vec_cbind(foobar(data.frame(x = 1)), data.frame(y = 2)),
     data.frame(x = 1, y = 2)
   )
-  expect_incompatible_df(
+  expect_equal(
     vec_rbind(foo, data.frame(), foo),
     df
   )
-
-  verify_errors({
-    foo <- structure(mtcars[1:3], class = c("foo", "data.frame"))
-    bar <- structure(mtcars[4:6], class = c("bar", "data.frame"))
-    baz <- structure(mtcars[7:9], class = c("baz", "data.frame"))
-    # Nested expect_warning() require testthat > 2.3.2
-    suppressWarnings(expect_warning(vec_ptype_common_df_fallback(foo, bar, baz)))
-    suppressWarnings(expect_warning(vec_ptype_common_df_fallback(foo, baz, bar, baz, foo, bar)))
-
-    with_fallback_warning(expect_df_fallback_warning(invisible(vec_rbind(foo, data.frame(), foo))))
-    with_fallback_warning(expect_df_fallback_warning(invisible(vec_cbind(foo, data.frame(x = 1)))))
-    with_fallback_warning(expect_df_fallback_warning(invisible(vec_cbind(foo, data.frame(x = 1), bar))))
-
-    with_fallback_quiet(invisible(vec_rbind(foo, data.frame(), foo)))
-    with_fallback_quiet(invisible(vec_cbind(foo, data.frame(x = 1))))
-    with_fallback_quiet(invisible(vec_cbind(foo, data.frame(x = 1), bar)))
-  })
 })
 
 
@@ -233,6 +196,18 @@ test_that("casting to and from data frame preserves row names", {
   expect_identical(row.names(out), row.names(mtcars))
 })
 
+test_that("df_cast() evaluates arg lazily", {
+  expect_silent(df_cast(data_frame(), data_frame(), x_arg = print("oof")))
+  expect_silent(df_cast(data_frame(), data_frame(), to_arg = print("oof")))
+})
+
+# df_ptype2 ---------------------------------------------------------------
+
+test_that("df_ptype2() evaluates arg lazily", {
+  expect_silent(df_ptype2(data_frame(), data_frame(), x_arg = print("oof")))
+  expect_silent(df_ptype2(data_frame(), data_frame(), y_arg = print("oof")))
+})
+
 
 # new_data_frame ----------------------------------------------------------
 
@@ -365,31 +340,46 @@ test_that("n and row.names (#894)", {
 })
 
 test_that("`x` must be a list", {
-  expect_error(new_data_frame(1), "`x` must be a list")
+  expect_snapshot((expect_error(
+    new_data_frame(1),
+    "`x` must be a list"
+  )))
 })
 
 test_that("if supplied, `n` must be an integer of size 1", {
-  expect_error(new_data_frame(n = c(1L, 2L)), "must be an integer of size 1")
-  expect_error(new_data_frame(n = "x"), "must be an integer of size 1")
+  expect_snapshot({
+    (expect_error(new_data_frame(n = c(1L, 2L)), "must be an integer of size 1"))
+    (expect_error(new_data_frame(n = "x"), "must be an integer of size 1"))
+  })
+})
+
+test_that("if supplied, `n` can't be negative or missing (#1477)", {
+  expect_snapshot({
+    (expect_error(new_data_frame(n = -1L)))
+    (expect_error(new_data_frame(n = NA_integer_)))
+  })
 })
 
 test_that("`class` must be a character vector", {
-  expect_error(new_data_frame(class = 1), "must be NULL or a character vector")
+  expect_snapshot((expect_error(
+    new_data_frame(class = 1),
+    "must be NULL or a character vector"
+  )))
 })
 
-test_that("flat width is computed", {
-  df_flat_width <- function(x) {
-    .Call(vctrs_df_flat_width, x)
+test_that("flatten info is computed", {
+  df_flatten_info <- function(x) {
+    .Call(ffi_df_flatten_info, x)
   }
-  expect_identical(df_flat_width(mtcars), ncol(mtcars))
+  expect_identical(df_flatten_info(mtcars), list(FALSE, ncol(mtcars)))
 
   df <- tibble(x = 1, y = tibble(x = 2, y = tibble(x = 3), z = 4), z = 5)
-  expect_identical(df_flat_width(df), 5L)
+  expect_identical(df_flatten_info(df), list(TRUE, 5L))
 })
 
 test_that("can flatten data frames", {
   df_flatten <- function(x) {
-    .Call(vctrs_df_flatten, x)
+    .Call(ffi_df_flatten, x)
   }
   expect_identical(df_flatten(mtcars), mtcars)
 
@@ -397,6 +387,16 @@ test_that("can flatten data frames", {
   expect_identical(df_flatten(df), new_data_frame(list(x = 1, x = 2, x = 3, z = 4, z = 5)))
 })
 
+test_that("can flatten data frames with rcrd columns (#1318)", {
+  col <- new_rcrd(list(a = 1))
+  df <- data_frame(col = col, y = 1)
+  expect_identical(vec_proxy_equal(df), data_frame(col = 1, y = 1))
+
+  col <- new_rcrd(list(a = 1, b = 2))
+  df <- data_frame(col = col, y = 1)
+  expect_identical(vec_proxy_equal(df), data_frame(a = 1, b = 2, y = 1))
+})
+
 test_that("new_data_frame() zaps existing attributes", {
   struct <- structure(list(), foo = 1)
   expect_identical(
@@ -411,6 +411,20 @@ test_that("new_data_frame() zaps existing attributes", {
 
 # data_frame --------------------------------------------------------------
 
+test_that("data_frame() and df_list() report error context", {
+  expect_snapshot({
+    (expect_error(data_frame(a = 1, a = 1)))
+    (expect_error(data_frame(a = 1, a = 1, .error_call = call("foo"))))
+    (expect_error(data_frame(a = 1:2, b = int())))
+    (expect_error(data_frame(a = 1:2, b = int(), .error_call = call("foo"))))
+
+    (expect_error(df_list(a = 1, a = 1)))
+    (expect_error(df_list(a = 1, a = 1, .error_call = call("foo"))))
+    (expect_error(df_list(a = 1:2, b = int())))
+    (expect_error(df_list(a = 1:2, b = int(), .error_call = call("foo"))))
+  })
+})
+
 test_that("can construct data frames with empty input", {
   expect_identical(data_frame(), new_data_frame())
   expect_named(data_frame(), character())
@@ -427,7 +441,9 @@ test_that("input is tidy recycled", {
     data_frame(x = double(), y = integer())
   )
 
-  expect_error(data_frame(1:2, 1:3), class = "vctrs_error_incompatible_size")
+  expect_snapshot({
+    expect_error(data_frame(1:2, 1:3), class = "vctrs_error_incompatible_size")
+  })
 })
 
 test_that("dots are dynamic", {
@@ -443,25 +459,52 @@ test_that("unnamed input is auto named with empty strings", {
   expect_named(data_frame(1, 2, .name_repair = "minimal"), c("", ""))
 })
 
-test_that("unnamed data frames are auto spliced", {
+test_that("unnamed data frames are auto unpacked", {
   expect_identical(
     data_frame(w = 1, data_frame(x = 2, y = 3), z = 4),
     data_frame(w = 1, x = 2, y = 3, z = 4)
   )
 })
 
-test_that("named data frames are not spliced", {
+test_that("named data frames are not unpacked", {
   df_col <- data_frame(x = 2, y = 3)
   df <- data_frame(w = 1, col = data_frame(x = 2, y = 3), z = 4)
 
   expect_identical(df$col, df_col)
 })
 
-test_that("spliced data frames without names are caught", {
+test_that("unpacked data frames without names are caught", {
   df_col <- new_data_frame(list(1))
   expect_error(data_frame(df_col), "corrupt data frame")
 })
 
+test_that("unpacking in `df_list()` can be disabled with `.unpack = FALSE`", {
+  out <- df_list(
+    w = 1,
+    data_frame(x = 2, y = 3),
+    z = 4,
+    .unpack = FALSE,
+    .name_repair = "minimal"
+  )
+
+  expect <- list(
+    w = 1,
+    data_frame(x = 2, y = 3),
+    z = 4
+  )
+
+  expect_identical(out, expect)
+})
+
+test_that("`.unpack` is validated", {
+  expect_snapshot(error = TRUE, {
+    df_list(.unpack = 1)
+  })
+  expect_snapshot(error = TRUE, {
+    df_list(.unpack = c(TRUE, FALSE))
+  })
+})
+
 test_that("`NULL` inputs are dropped", {
   expect_identical(data_frame(NULL, x = 1, NULL), data_frame(x = 1))
 })
@@ -481,24 +524,34 @@ test_that("`.size` can force a desired size", {
 })
 
 test_that("`.name_repair` repairs names", {
-  expect_named(
-    expect_message(data_frame(x = 1, x = 1, .name_repair = "unique")),
-    c("x...1", "x...2")
-  )
+  expect_message(res <- data_frame(x = 1, x = 1, .name_repair = "unique"))
+  expect_named(res, c("x...1", "x...2"))
 })
 
 test_that("`.name_repair` happens after auto-naming with empty strings", {
-  expect_named(
-    expect_message(data_frame(1, 2, .name_repair = "unique")),
-    c("...1", "...2")
-  )
+  expect_message(res <- data_frame(1, 2, .name_repair = "unique"))
+  expect_named(res, c("...1", "...2"))
 })
 
 test_that("`.name_repair` happens after splicing", {
-  expect_named(
-    expect_message(data_frame(x = 1, data_frame(x = 2), .name_repair = "unique")),
-    c("x...1", "x...2")
-  )
+  expect_message(res <- data_frame(x = 1, data_frame(x = 2), .name_repair = "unique"))
+  expect_named(res, c("x...1", "x...2"))
+})
+
+test_that("`.name_repair` can be quiet", {
+  local_name_repair_verbose()
+
+  expect_snapshot({
+    dfl_unique <- df_list(1, 2, .name_repair = "unique_quiet")
+    dfl_universal <- df_list("if" = 1, "in" = 2, .name_repair = "universal_quiet")
+    df_unique <- data_frame(1, 2, .name_repair = "unique_quiet")
+    df_universal <- data_frame("if" = 1, "in" = 2, .name_repair = "universal_quiet")
+  })
+
+  expect_named(dfl_unique, c("...1", "...2"))
+  expect_named(dfl_universal, c(".if", ".in"))
+  expect_named(df_unique, c("...1", "...2"))
+  expect_named(df_universal, c(".if", ".in"))
 })
 
 # fallback ----------------------------------------------------------------
@@ -512,23 +565,25 @@ test_that("data frame fallback handles column types (#999)", {
   expect_identical(vec_ptype2(df1, df2), common)
   expect_identical(vec_ptype2(df2, df1), common)
 
-  expect_error(
-    vec_ptype2(df1, df3),
-    class = "vctrs_error_incompatible_type"
-  )
-  expect_error(
-    vec_ptype2(df3, df1),
-    class = "vctrs_error_incompatible_type"
-  )
+  exp <- foobar(data_frame(x = 1, y = na_dbl))
+  expect_identical(vec_cast(df1, df2), exp)
+  expect_identical(vec_cast(set_tibble(df1), set_tibble(df2)), set_tibble(exp))
 
-  expect_identical(
-    vec_cast(df1, df2),
-    foobar(data.frame(x = 1, y = na_dbl))
-  )
-  expect_error(
-    vec_cast(df2, df1),
-    class = "vctrs_error_cast_lossy"
-  )
+  expect_snapshot({
+    local_error_call(call("my_function"))
+    (expect_error(
+      vec_ptype2(df1, df3),
+      class = "vctrs_error_incompatible_type"
+    ))
+    (expect_error(
+      vec_ptype2(df3, df1),
+      class = "vctrs_error_incompatible_type"
+    ))
+    (expect_error(
+      vec_cast(df2, df1),
+      class = "vctrs_error_incompatible_type"
+    ))
+  })
 
   expect_identical(
     vec_rbind(df1, df2),
@@ -539,7 +594,7 @@ test_that("data frame fallback handles column types (#999)", {
   df1_attrib <- foobar(df1, foo = "foo")
   df2_attrib <- foobar(df2, bar = "bar")
   exp <- data.frame(x = c(1, 1), y = c(NA, 2))
-  expect_incompatible_df(
+  expect_equal(
     vec_rbind(df1_attrib, df2_attrib),
     exp
   )
@@ -554,59 +609,14 @@ test_that("data frame fallback handles column types (#999)", {
     },
     vec_rbind(df1_attrib, df2_attrib)
   )
-
-  expect_identical(out, foobar(exp, dispatched = TRUE))
+  expect_identical(out, exp)
 })
 
 test_that("falls back to tibble for tibble subclasses (#1025)", {
   foo <- foobar(tibble::as_tibble(mtcars))
-  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars)), "tbl_df")
-  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, mtcars)), "tbl_df")
-  expect_s3_class(expect_df_fallback_warning_maybe(vec_rbind(foo, mtcars, foobar(mtcars))), "tbl_df")
-
-  verify_errors({
-    with_fallback_warning(expect_df_fallback_warning(
-      vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobaz(mtcars)
-      )
-    ))
-    with_fallback_warning(expect_df_fallback_warning(
-      vec_rbind(
-        tibble::as_tibble(mtcars),
-        foobar(tibble::as_tibble(mtcars))
-      )
-    ))
-    with_fallback_warning(expect_df_fallback_warning(
-      vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobar(tibble::as_tibble(mtcars))
-      )
-    ))
-
-    with_fallback_quiet(
-      vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobaz(mtcars)
-      )
-    )
-    with_fallback_quiet(
-      vec_rbind(
-        tibble::as_tibble(mtcars),
-        foobar(tibble::as_tibble(mtcars))
-      )
-    )
-    with_fallback_quiet(
-      vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobar(tibble::as_tibble(mtcars))
-      )
-    )
-  })
+  expect_s3_class(vec_rbind(foo, mtcars), "tbl_df")
+  expect_s3_class(vec_rbind(foo, mtcars, mtcars), "tbl_df")
+  expect_s3_class(vec_rbind(foo, mtcars, foobar(mtcars)), "tbl_df")
 })
 
 test_that("fallback is recursive", {
@@ -617,75 +627,8 @@ test_that("fallback is recursive", {
   baz <- new_data_frame(list(y = 1:3, x = foobar(df, bar = TRUE)))
 
   exp <- new_data_frame(list(x = vec_rbind(df, df)))
-  expect_incompatible_df(vec_rbind(foo, bar), exp)
+  expect_equal(vec_rbind(foo, bar), exp)
 
   exp <- new_data_frame(list(x = vec_rbind(df, df), y = c(NA, NA, NA, 1:3)))
-  expect_incompatible_df(vec_rbind(foo, baz), exp)
-})
-
-test_that("data frame output is informative", {
-  verify_output(test_path("error", "test-type-data-frame.txt"), {
-    "# combining data frames with foreign classes uses fallback"
-    foo <- structure(mtcars[1:3], class = c("foo", "data.frame"))
-    bar <- structure(mtcars[4:6], class = c("bar", "data.frame"))
-    baz <- structure(mtcars[7:9], class = c("baz", "data.frame"))
-    vec_ptype_common_df_fallback(foo, bar, baz)
-    vec_ptype_common_df_fallback(foo, baz, bar, baz, foo, bar)
-
-    with_fallback_warning(invisible(vec_rbind(foo, data.frame(), foo)))
-    with_fallback_warning(invisible(vec_rbind(foo, baz, bar, baz, foo, bar)))
-
-    with_fallback_warning(invisible(vec_cbind(foo, data.frame(x = 1))))
-    with_fallback_warning(invisible(vec_cbind(foo, data.frame(x = 1), bar)))
-
-    with_fallback_quiet(invisible(vec_rbind(foo, data.frame(), foo)))
-    with_fallback_quiet(invisible(vec_rbind(foo, baz, bar, baz, foo, bar)))
-
-    with_fallback_quiet(invisible(vec_cbind(foo, data.frame(x = 1))))
-    with_fallback_quiet(invisible(vec_cbind(foo, data.frame(x = 1), bar)))
-
-    "# falls back to tibble for tibble subclasses (#1025)"
-    with_fallback_warning(
-      invisible(vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobaz(mtcars)
-      ))
-    )
-    with_fallback_warning(
-      invisible(vec_rbind(
-        tibble::as_tibble(mtcars),
-        foobar(tibble::as_tibble(mtcars))
-      ))
-    )
-    with_fallback_warning(
-      invisible(vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobar(tibble::as_tibble(mtcars))
-      ))
-    )
-
-    with_fallback_quiet(
-      invisible(vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobaz(mtcars)
-      ))
-    )
-
-    with_fallback_quiet(
-      invisible(vec_rbind(
-        tibble::as_tibble(mtcars),
-        foobar(tibble::as_tibble(mtcars))
-      ))
-    )
-    with_fallback_quiet(
-      invisible(vec_rbind(
-        foobar(tibble::as_tibble(mtcars)),
-        mtcars,
-        foobar(tibble::as_tibble(mtcars))
-      ))
-    )
-  })
+  expect_equal(vec_rbind(foo, baz), exp)
 })
